/****************************************************************
*								
*		    Particle Member functions			
*			by Pran Mukherjee
*			pran@umich.edu
*			http://www.erim-int.com/~pran
*			04/15/00
*
****************************************************************/
#include "ParticleAPI.h"
#include "Particle.h"

#include "Smoke.h"
#include "Contrail.h"
#include "Explosion.h"
#include "Firework.h"

// CONSTRUCTORS
Particle::Particle(float pos[3], float vel[3], float col[4], float size, float mass, float fade, int keep)

{
	P_Pos[X] = pos[X];
	P_Pos[Y] = pos[Y];
	P_Pos[Z] = pos[Z];
	P_Vel[X] = vel[X];
	P_Vel[Y] = vel[Y];
	P_Vel[Z] = vel[Z];
	P_Color[R] = col[R];
	P_Color[G] = col[G];
	P_Color[B] = col[B];
	P_Color[A] = col[A];
	P_Size = size;
	P_Mass = mass;
	P_Fade_Rate = fade;
    next = NULL;
	prev = NULL;
	S_Parent = NULL;
	RS_Parent = NULL;
	s = NULL;
	A_Type = NONE;						// By default NO particle is an anchor.  Only becomes so when attached to RootSystem
	keep_systems = keep;
	//    fprintf(stderr, "Called default Particle constructor\n");
}

// DESTRUCTORS
Particle::~Particle()
{
	ParticleSys *t, *c;

	if ((NULL != S_Parent) || (NULL != RS_Parent)){
		fprintf(stderr, "Particle is still attached to a ParticleSys or RootSystem.  Failing.\n");
		exit(1);
	}

	if (next || prev) {
		fprintf(stderr, "Particle is still connected to previous or subsequent particles (not properly detached).  Failing.\n");
		exit(1);
	}

	// Kill all ParticleSystems attached to Particle
	c = s;
	while (NULL != c)
	{
		t = c;
		c = c->next;
		P_DeleteSystem(t);
		free(t);
	}
	s = NULL;    
//	fprintf(stderr, "Called default Particle destructor\n");
}

// METHODS
Particle::P_AttachSystem(ParticleSys *newSys)
{
	ParticleSys *c;
	c = s;

	if (newSys->P_Parent || newSys->RS_Parent) {
		fprintf(stderr, "System is already attached to another particle or RootSystem system.  Failing.\n");
		return(false);
	}

	if (newSys->next || newSys->prev) {
		fprintf(stderr, "System is still connected to previous or subsequent systems (not properly detached).  Failing.\n");
		return(false);
	}

	if (NULL != c)
	{
		while (c->next)			// Find end of list
		{
			c = c->next;
		}
		c->next = newSys;				//Append new system to end of list
		c->next->prev = c;				// Set new system's backward link to former end of list
	} 
	else								// System is first in list
	{
		s = newSys;
		newSys->prev = NULL;
	}
	newSys->P_Parent = this;
	newSys->RS_Parent = NULL;
	return(true);
}

Particle::P_DetachSystem(ParticleSys *detSys)
{
	if (NULL == detSys->P_Parent) return(false);
	if (detSys->RS_Parent) {
		fprintf(stderr, "This system is attached to a RootSystem system. You made a mistake.\n");
		return(false);
	}
	if (detSys->next)			// Not end of set 
		if (detSys->prev)		// Middle of set
		{
			detSys->next->prev = detSys->prev;		
			detSys->prev->next = detSys->next;
		}		
		else 								// Top of set attached to Particle
		{
			detSys->next->prev = NULL;
			s = detSys->next;	
		}
	else if (detSys->prev)	// End of set
	{
		detSys->prev->next = NULL;
	}
	else		// Only system attached
		{
		s = NULL;
	}

	detSys->P_Parent = NULL;
	detSys->next = NULL;
	detSys->prev = NULL;
	return(true);
}

Particle::P_DeleteSystem(ParticleSys *killSys)
{
	P_DetachSystem(killSys);
	killSys->~ParticleSys();
	return(true);
}

Particle::P_SetType(Anchor_Type a_type)
{
	switch(a_type) {
	case SMOKE:
	case FLAME:
	case SHOT:
	case EXPLOSION:
	case FIREWORK:
		A_Type = a_type;
		fprintf(stderr, "Anchor set to type %d\n", A_Type);
		break;
	default:
		fprintf(stderr, "The type specified for this anchor is not allowed.  Failing.\n");
		return(false);
	}
	return(true);
}

Particle::P_RunParticle(float accel[3], float force[3], float dt, PS_Type type, PS_Modes out_mode)
{
	ParticleSys *c = NULL;

	// Draw the particle
	switch (out_mode.o_type){
	case OGL_PT:
		P_DrawPoint();
		break;
	case OGL_QUAD:
		P_DrawQuad(out_mode.image);
		break;
	default:
		fprintf(stderr, "Invalid output mode\n");
	}

	// Draw all systems attached to particle
	c = s;
	while (NULL != c)
	{
		c->S_RunSystem(accel, force, dt);				// TEMPORARY;  FIX THIS
		c = c->next;
	}

	if ((NONE == A_Type) && (NONE != type)) 
	{
		P_UpdateParticle(accel, force, dt, type);	// Update particle and kill it if necessary
	}
	else if ((NONE != A_Type) && (NONE == type))
	{
		P_UpdateAnchor(accel, force, dt);			// Update anchor and kill it if necessary
	}
	else
	{
		fprintf(stderr, "There is neither anchor type nor system type. Unable to update particle. Failing.\n");
		exit(1);
	}
	return(true);
}

Particle::P_UpdateParticle(float accel[3], float force[3], float dt, PS_Type type)
{
	P_Color[A] -= P_Fade_Rate * dt;

	if (P_Color[A] < 0.0f)			// Burnt-out particle
	{
		if ((NULL != s) && (keep_systems))  		// Move any Systems to Parent system if this particle is set to do so ( mostly shots or fireworks)
		{
			ParticleSys *tmp = s;
			while (tmp) {
				P_DetachSystem(tmp);
				if (S_Parent) S_Parent->S_AttachSystem(tmp);
				else if (RS_Parent) RS_Parent->RS_AttachSystem(tmp);
				else
				{
					fprintf (stderr, "System attached to neither RootSystem nor ParticleSys. Failing.\n");
					exit(1);
				}	// End if
				tmp = s;
			}		// End while
		}			// End if  systems
	if (S_Parent) S_Parent->S_DeleteParticle(this);		// Kill dead particle
	else if (RS_Parent) RS_Parent->RS_DeleteAnchor(this);		
	return(true);
	}				// End if dead

	// If particle still lives, move/accelerate it
	switch (type) {					// System-type specific effects occur here
	case SMOKE:
		P_Size += 3 * dt;
		P_Pos[X] += (P_Vel[X] * dt);
		P_Pos[Y] += (P_Vel[Y] * dt);
		P_Pos[Z] += (P_Vel[Z] * dt);

		if (P_Vel[Y] > 1.0f) P_Vel[Y] -= dt;		// Vertical acceleration -1.0m/s^2
		break;
	case CONTRAIL:
		P_Size += 3 * dt;
		P_Pos[X] += (P_Vel[X] * dt);
		P_Pos[Y] += (P_Vel[Y] * dt);
		P_Pos[Z] += (P_Vel[Z] * dt);

		if ((P_Vel[X] > 0.2f) || (P_Vel[X] < -0.2f)) P_Vel[X] *=  0.9f;		// Velocity SHOULD halfs each second, then stays really low
		else if (P_Vel[X] > 0) P_Vel[X] = 0.2f;
		else P_Vel[X] = -0.2f;											

		if ((P_Vel[Y] > 0.2f) || (P_Vel[Y] < -0.2f)) P_Vel[Y] *= 0.9f;		
		else if (P_Vel[Y] > 0) P_Vel[Y] = 0.2f;
		else P_Vel[Y] = -0.2f;											

		if ((P_Vel[Z] > 0.2f) || (P_Vel[Z] < -0.2f)) P_Vel[Z] *=  0.9f;	
		else if (P_Vel[Z] > 0) P_Vel[Z] = 0.2f;
		else P_Vel[Z] = -0.2f;											
		break;

	case FIREWORK:
		if (P_Color[A] <= 0.5)				// Sparks now spawn
		{
			P_Color[A] = -1.0;  // This will kill the particle next iteration and move its systems to its parent
			float t_vel[3] = {0.0f, 0.0f, 0.0f};
			Explosion *temp;
			temp = new Explosion(100, 0, P_Pos, t_vel);
			temp->S_InitExplosion(dt, 5.0);		
			if (RS_Parent)
			{
				RS_Parent->RS_AttachSystem(temp);
			}
			else if (this->S_Parent)
			{
				S_Parent->S_AttachSystem(temp);
			}
		}

		P_Pos[X] += (P_Vel[X] *dt + accel[X] * dt * dt);				// Should be P_Pos += (P_Vel * dt + P_Accel * dt^2)  I think
		P_Pos[Y] += (P_Vel[Y]  *dt + accel[Y] * dt * dt);
		P_Pos[Z] += (P_Vel[Z]  *dt + accel[Z] * dt * dt);
	
		P_Vel[X] += accel[X] * dt;				// Should be P_Vel += P_Vel * dt;
		P_Vel[Y] += accel[Y] * dt;
		P_Vel[Z] += accel[Z] * dt;

		break;

	default:
		P_Pos[X] += (P_Vel[X] * dt + 1/2 * accel[X] * dt * dt);				
		P_Pos[Y] += (P_Vel[Y] * dt + 1/2 * accel[Y] * dt * dt);
		P_Pos[Z] += (P_Vel[Z] * dt + 1/2 * accel[Z] * dt * dt);

		P_Vel[X] += accel[X] * dt;				
		P_Vel[Y] += accel[Y] * dt;
		P_Vel[Z] += accel[Z] * dt;
	}

	/*
	Add forces vs mass to accelerate again
	*/
return(true);
}

Particle::P_UpdateAnchor(float accel[3], float force[3], float dt)
{
	P_Color[A] -= P_Fade_Rate * dt;

	if (P_Color[A] < 0.0f)			// Burnt-out particle
	{
		if ((NULL != s) && (keep_systems))  		// Move any Systems to Parent system if this particle is set to do so ( mostly shots or fireworks)
		{
			ParticleSys *tmp = s;
			while (tmp) {
				P_DetachSystem(tmp);
				if (S_Parent) S_Parent->S_AttachSystem(tmp);
				else if (RS_Parent) 
				{
					RS_Parent->RS_AttachSystem(tmp);
				}
				else
				{
					fprintf (stderr, "System attached to neither RootSystem nor ParticleSys. Failing.\n");
					exit(1);
				}
				tmp = s;
			}	// End while
		}		// End if
		if (S_Parent) S_Parent->S_DeleteParticle(this);		// Kill dead particle
		else if (RS_Parent) 
		{
			RS_Parent->RS_DeleteAnchor(this);		
		}
	return(true);
	}			// End if dead

	// If particle still lives, move/accelerate it

	switch (A_Type) {					// System-type specific effects occur here
	case SHOT:
		if (P_Pos[Y] <= 0.0)				// Shell hit or passed the ground
		{
			float t_vel[3] = {0.0f, 0.0f, 0.0f}, s_vel[3] = {0.0f, 6.0f, 0.0f}; 
			Explosion *expl = new Explosion(1000, 0, P_Pos, t_vel);
			Smoke *smk = new Smoke(500, 15.0f, P_Pos, t_vel, s_vel);

			expl->S_InitExplosion(dt, 15.0);			// Lots of fuel
			P_AttachSystem(expl);
			P_AttachSystem(smk);

			P_Color[A] = -1.0;  // This will kill the particle next iteration and move all systems to its parent
		}

		if (NULL == s)				// This may be a bad idea.  System can't live without contrail.  (?)
		{
			RS_Parent->RS_DeleteAnchor(this);
		}

		P_Pos[X] += (P_Vel[X] *dt + accel[X] * dt * dt);				// Should be P_Pos += (P_Vel * dt + P_Accel * dt^2)  I think
		P_Pos[Y] += (P_Vel[Y]  *dt + accel[Y] * dt * dt);
		P_Pos[Z] += (P_Vel[Z]  *dt + accel[Z] * dt * dt);
	
		P_Vel[X] += accel[X] * dt;				// Should be P_Vel += P_Vel * dt;
		P_Vel[Y] += accel[Y] * dt;
		P_Vel[Z] += accel[Z] * dt;
		break;

	case FIREWORK:
		if (P_Vel[Y] <= 5.0)				// Shell reaches peak of arc
		{
			Firework *burst = new Firework(10, P_Pos, P_Vel);
			burst->S_InitFirework(35.0);			// Lots of fuel
			P_AttachSystem(burst);

			P_Color[A] = -1.0;  // This will kill the particle next iteration and move its systems to its parent
		}

		if (NULL == s) 
		{
			RS_Parent->RS_DeleteAnchor(this);
		}

		P_Pos[X] += (P_Vel[X] *dt + accel[X] * dt * dt);				
		P_Pos[Y] += (P_Vel[Y]  *dt + accel[Y] * dt * dt);
		P_Pos[Z] += (P_Vel[Z]  *dt + accel[Z] * dt * dt);
	
		P_Vel[X] += accel[X] * dt;				
		P_Vel[Y] += accel[Y] * dt;
		P_Vel[Z] += accel[Z] * dt;
		break;
}

	/*
	Move and accelerate particle, then run forces vs mass to accelerate again
	*/
return(true);
}

Particle::P_DrawPoint()
{
	glPointSize(P_Size);
	glBegin(GL_POINTS);
		glColor4fv(P_Color);
		glVertex3fv(P_Pos);
	glEnd();
	return(true);
}

Particle::P_DrawQuad(GLuint image)
{
	glPushMatrix();
//		glLoadIdentity();								// Reset The ModelView Matrix
		glTranslatef(P_Pos[X], P_Pos[Y], P_Pos[Z]);
		glColor4fv(P_Color);
		glCallList(image);									// Draw Our Particle
	glPopMatrix();
	return(true);
}

