/****************************************************************
*								
*		    ParticleSys Member functions			
*			by Pran Mukherjee
*			pran@umich.edu
*			http://www.erim-int.com/~pran
*			04/15/00
*								
****************************************************************/
#include "ParticleAPI.h"
#include "ParticleSys.h"
#include "Vectors.h"

friend class ParticleSys;
friend class RootSystem;

// CONSTRUCTORS
ParticleSys::ParticleSys(int max_p, float pos[3], float vel[3], PS_Type type)
{
	S_Pos[X] = pos[X];
	S_Pos[Y] = pos[Y];
	S_Pos[Z] = pos[Z];
	S_Vel[X] = vel[X];
	S_Vel[Y] = vel[Y];
	S_Vel[Z] = vel[Z];
	S_Num_Particles = 0;
	S_Max_Particles = max_p;
	S_Current_Max_P = (float)max_p;
	S_Type = type;
	p = NULL;
	next = NULL;
	prev = NULL;
	P_Parent = NULL;
	RS_Parent = NULL;
	sys_out_mode.o_type = OGL_QUAD;
	MakeParticleQuad();
//    fprintf(stderr, "Called ParticleSys constructor\n");
}

// DESTRUCTOR
ParticleSys::~ParticleSys()
{
	Particle *t, *c;

	if (P_Parent || RS_Parent) {
		fprintf(stderr, "System is still attached to another RootSystem system or particle.  Failing.\n");
		exit(1);
	}

	if (next || prev) {
		fprintf(stderr, "System is still connected to previous or subsequent systems (not properly detached).  Failing.\n");
		exit(1);
	}

	// Kill all particles attached to ParticleSys
	c = p;
	while (NULL != c)
	{
		t = c;
		c = c->next;
		S_DeleteParticle(t);
		free(t);
	}
	p = NULL;    
	
   fprintf(stderr, "Called default ParticleSys destructor\n");
}

// METHODS
ParticleSys::S_AddParticle(float p_tmp[3], float v_tmp[3], float c_tmp[4], float size, float mass, float fade, int keep_systems)													
{
	if ((float)S_Num_Particles >= S_Current_Max_P){
		return (false);
	} 
	else 
	{
		Particle *n = new Particle(p_tmp, v_tmp, c_tmp, size, mass, fade, keep_systems);
		S_AttachParticle(n);
	}
	return(true);
}

ParticleSys::S_AttachParticle(Particle *newPart)
{
	Particle *c = p;

	if (newPart->S_Parent || newPart->RS_Parent) {
		fprintf(stderr, "Particle is already attached to another RootSystem or ParticleSys.  Failing.\n");
		return(false);
	}

	if (newPart->next || newPart->prev) {
		fprintf(stderr, "Particle is still connected to previous or subsequent particles (not properly detached).  Failing.\n");
		return(false);
	}

	if (NULL != c)
	{
		while (c->next) 
		{
			c = c->next;
		}
		c->next = newPart;				//Append particle to end of list
		newPart->prev = c;				// Set new particle's backward link to former end of list
	} 
	else								// Particle is first in list
	{
		p = newPart;
		p->prev = NULL;
	}
	newPart->S_Parent = this;
	newPart->RS_Parent = NULL;
	S_Num_Particles++;
	return(true);
}

ParticleSys::S_DetachParticle(Particle *detPart)
{
	if(NULL == detPart->S_Parent) return(false);
	if (detPart->RS_Parent) {
		fprintf(stderr, "This particle is attached to a RootSystem. You made a mistake.\n");
		return(false);
	}

	if (detPart->next)			// Not end of set 
		if (detPart->prev)		// Middle of set
		{
			detPart->next->prev = detPart->prev;		
			detPart->prev->next = detPart->next;
		}		
		else								// Top of list
		{
			detPart->next->prev = NULL;
			p = detPart->next;
		}
	else if (detPart->prev)			// End of set
		detPart->prev->next = NULL;
	else		// Only particle attached
		p = NULL;

	S_Num_Particles --;
	detPart->S_Parent = NULL;
	detPart->next = NULL;
	detPart->prev = NULL;
	return(true);
}

ParticleSys::S_DeleteParticle(Particle *killPart)
{
	S_DetachParticle(killPart);
	killPart->~Particle();
	return(true);
 }

ParticleSys::S_AttachSystem(ParticleSys *newSys)
{
	ParticleSys *c;
	c = this;

	if (newSys->P_Parent || newSys->RS_Parent) {
		fprintf(stderr, "System is already attached to a particle or RootSystem system.  Failing.\n");
		return(false);
	}

	if (newSys->next || newSys->prev) {
		fprintf(stderr, "System is still connected to previous or subsequent systems (not properly detached).  Failing.\n");
		return(false);
	}

	while (NULL != c->next)  c = c->next;		// Find end of system list
	c->next = newSys;				//Append new system to end of list
	c->next->prev = c;				// Set new system's backward link to former end of list

	newSys->P_Parent = this->P_Parent;				// Set P_Parent and RS_Parent to same as THIS
	newSys->RS_Parent = this->RS_Parent;
	return(true);
}

ParticleSys::S_UpdateSystemFlame()
{
	return(true);
}

ParticleSys::S_UpdateSystem(float dt)				//  NOTE:  Just a placeholder for virtual function
{
	return(true);
}
ParticleSys::S_RunSystem(float accel[3], float force[3], float dt) 
{
	Particle *c = p;

	if (OGL_QUAD == sys_out_mode.o_type)
	{
		glShadeModel(GL_SMOOTH);							// Enable Smooth Shading
		glClearColor(0.0f,0.0f,0.0f,0.0f);					// Black Background
		glClearDepth(1.0f);									// Depth Buffer Setup
		glDisable(GL_DEPTH_TEST);							// Disable Depth Testing
		glEnable(GL_BLEND);									// Enable Blending
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);					// Type Of Blending To Perform
		glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);	// Really Nice Perspective Calculations
		glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);				// Really Nice Point Smoothing
		glEnable(GL_TEXTURE_2D);							// Enable Texture Mapping
	
		glBindTexture(GL_TEXTURE_2D, sys_out_mode.texture[0]);
	}
	else
	{
		glClearColor(0.0f,0.0f,0.0f,0.0f);					// Black Background
		glClearDepth(1.0f);									// Depth Buffer Setup
		glDisable(GL_DEPTH_TEST);							// Disable Depth Testing
		glEnable(GL_BLEND);									// Enable Blending
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);					// Type Of Blending To Perform
	}

	while (NULL != c)
	{
		c->P_RunParticle(accel, force, dt, S_Type, sys_out_mode);		
		c = c->next;
	}

	// Move Systems
	if ((NULL != RS_Parent) && (NULL == P_Parent))	// System is attached to RootSystem
	{
		//Move and accelerate (no use of Forces) system, but NOT particles in it
	}
	else if ((NULL == RS_Parent) && (NULL != P_Parent))  // System is attached to Particle
	{
		// By definition its parent particle MUST have already been moved and accelerated;
		S_Pos[X] = P_Parent->P_Pos[X];	// So that the WHOLE system stays attached to particle
		S_Pos[Y] = P_Parent->P_Pos[Y];	
		S_Pos[Z] = P_Parent->P_Pos[Z];	
		S_Vel[X] = P_Parent->P_Vel[X];	
		S_Vel[Y] = P_Parent->P_Vel[Y];	
		S_Vel[Z] = P_Parent->P_Vel[Z];	
	}
	else
	{
		fprintf (stderr, "S_RunSystem: System cannot be moved since it's not attached to anything.\n");
	}

	switch (S_Type) {
	case SMOKE:
	case CONTRAIL:
	case EXPLOSION:
	case FIREWORK:
		S_UpdateSystem(dt);
		break;
	case FLAME:
		this->S_UpdateSystemFlame();
		break;
	case SHOT:
		fprintf(stderr, "This is NOT a system type; it's an anchor type.\n");
		exit(1);
		break;
	}

	if (OGL_QUAD == sys_out_mode.o_type)
	{
		glDisable(GL_TEXTURE_2D);							// Disable Texture Mapping
	}
	return(true);
}

ParticleSys::MakeParticleQuad()								// Build Our Particle Display List
{
	if (!LoadGLTextures(this->sys_out_mode.texture))								// Jump To Texture Loading Routine
	{
		exit (1);									// If Texture Didn't Load Return FALSE
	}
	this->sys_out_mode.image =glGenLists(1);								// Generate Our Display List
	glNewList(this->sys_out_mode.image,GL_COMPILE);						// Build A Compiled Display List
		glBegin(GL_QUADS);								// Build A Quad
			glTexCoord2f(1.0f,1.0f); glVertex3f( 0.5f, 0.5f, 0.0f);	// Top Right
			glTexCoord2f(0.0f,1.0f); glVertex3f(-0.5f, 0.5f, 0.0f);	// Top Left
			glTexCoord2f(0.0f,0.0f); glVertex3f(-0.5f,-0.5f, 0.0f);	// Bottom Left
			glTexCoord2f(1.0f,0.0f); glVertex3f( 0.5f,-0.5f, 0.0f);	// Bottom Right
		glEnd();										// Done Building Our Quad
	glEndList();										// Done Building Our Display List
}


AUX_RGBImageRec *LoadBMP(char *Filename)				// Loads A Bitmap Image
{
        FILE *File=NULL;								// File Handle
        if (!Filename)									// Make Sure A Filename Was Given
        {
                return NULL;							// If Not Return NULL
        }
        File=fopen(Filename,"r");						// Check To See If The File Exists
        if (File)										// Does The File Exist?
        {
			fclose(File);								// Close The Handle
			return auxDIBImageLoad(Filename);			// Load The Bitmap And Return A Pointer
        }
        return NULL;									// If Load Failed Return NULL
}

int LoadGLTextures(GLuint texture[1])									// Load Bitmap And Convert To Textures
{
	int Status=FALSE;								// Status Indicator
	AUX_RGBImageRec *TextureImage[1];				// Create Storage Space For The Textures
	memset(TextureImage,0,sizeof(void *)*1);		// Set The Pointer To NULL

	if (TextureImage[0]=LoadBMP("Data/Particle.bmp"))	// Load Particle Texture
	{
		Status=TRUE;								// Set The Status To TRUE
		glGenTextures(1, &texture[0]);				// Create One Texture

		glBindTexture(GL_TEXTURE_2D, texture[0]);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, 3, TextureImage[0]->sizeX, TextureImage[0]->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, TextureImage[0]->data);
    }
	else
	{
		fprintf(stderr, "Couldn't load file %s\n", "Data/Particle.bmp");
		exit(1);
	}
	if (TextureImage[0])							// If Texture Exists
	{
		if (TextureImage[0]->data)					// If Texture Image Exists
		{
			free(TextureImage[0]->data);			// Free The Texture Image Memory
		}
		free(TextureImage[0]);						// Free The Image Structure
	}
	return Status;									// Return The Status
}
